"""
streaming Kafka的处理、解析相关功能
"""

from pyspark.sql import functions as fn
from pyspark.sql.functions import col, get_json_object, from_unixtime, concat_ws, lit


# 根据需求，新请求从最新数据开始订阅。由于指定的group id，每次订阅将基于上次继续订阅。
def read_kafka(spark_ob, kafka_servers, kafka_topics, max_offset, unique_label):
    return spark_ob \
        .readStream \
        .format("kafka") \
        .option("kafka.bootstrap.servers", kafka_servers) \
        .option("kafka.group.id", unique_label) \
        .option("kafka.session.timeout.ms", 30000) \
        .option("subscribe", kafka_topics) \
        .option("maxOffsetsPerTrigger", f"{max_offset}") \
        .option("startingOffsets", "latest") \
        .option("failOnDataLoss", "false") \
        .option("includeHeaders", "false") \
        .load()


# 处理消息
def process_streaming(events, filter_project, filter_events, start_date, end_date):
    # 获取基础数据
    # 提取公用字段值并转换为string类型
    events = events.withColumn('partition', col('partition').cast('string'))  # 分区，转字符串
    events = events.withColumn('offset', col('offset').cast('string'))  # 分区内消息的偏移量，转字符串
    events = events.withColumn('data', col('value').cast('string'))  # 将二进制value转换为字符串
    events = events.withColumn('properties', get_json_object(events['data'].cast('string'), '$.properties'))  # 提取属性
    events = events.withColumn('project', get_json_object(events['data'], '$.project'))  # 提取project
    # 过滤项目（仅包含正式环境数据）
    events = events.filter(events['project'] == filter_project)  # 只保留正式项目数据
    events = events.withColumn('t', get_json_object(events['data'], '$.time'))  # 提取时间戳，字符串格式，临时公用

    # 加工以及提取特定字段，并输出原始streaming消息
    events = events.select(
        # 公用字段
        # 生成事件标识
        concat_ws('_', col('project'), col('topic'), col('partition'), col('offset')).alias('event_id'),
        # 项目_topic_分区_偏移量，保证唯一消息值
        get_json_object(events['data'], '$.event').alias('event'),  # 事件名
        # 获取不同格式的时间戳字段
        from_unixtime(events['t'] / 1000, "yyyy-MM-dd HH:mm:ss").cast('timestamp').cast('date').alias('date'),
        # 日期，date格式。注意使用datetime的时区
        from_unixtime(events['t'] / 1000, "yyyy-MM-dd HH:mm:ss").cast('timestamp').alias('datetime'),  # 时间戳，datetime格式
        events['t'].cast('bigint').alias('timestamp'),  # 时间戳，bigint格式

        # 提取用户各种ID，从(data)中获取
        get_json_object(events['data'], '$.user_id').alias('user_id'),  # cookieID
        get_json_object(events['data'], '$.login_id').alias('login_id'),  # AID
        get_json_object(events['data'], '$.device_id').alias('device_id'),  # 设备ID

        # 提取事件参数，从(properties)中获取
        get_json_object(events['properties'], '$.Eventcategory').alias('ec'),  # 事件类型，同GA的EC
        get_json_object(events['properties'], '$.Eventlabel').alias('el'),  # 事件标志，同GA的EL
        get_json_object(events['properties'], '$.Eventaction').alias('ea'),  # 事件动作，同GA的EA
        get_json_object(events['properties'], '$.$event_duration').alias('event_duration'),  # 事件时长

        # 提取页面和域信息，从properties)中获取
        get_json_object(events['properties'], '$.Appname').alias('app_name'),  # 应用名称，区分不同的平台
        get_json_object(events['properties'], '$.Hostname').alias('host_name'),  # 域名
        get_json_object(events['properties'], '$.Pagepath').alias('page_path'),  # 页面路径
        get_json_object(events['properties'], '$.Pagetitle').alias('page_title'),  # 页面名称

        # 提取可变json属性，方便后续扩展
        col('properties'),

        # 增加各种时间标记
        from_unixtime(events['t'] / 1000, "yyyy-MM-dd HH:mm:ss").cast('timestamp').alias('process_datetime'),
        # kafka消息时间，该时间不等于消息中的行为时间 timestamp
        fn.current_timestamp().alias('insert_datetime')  # 系统处理时间，约等于入库时间

    )

    # 过滤event 以及 日期内的数据
    return events.where(~col('event').isin(filter_events)).filter(col('date') >= (lit(start_date)).cast('date')).filter(
        col('date') <= (lit(end_date)).cast('date'))


# 写库
def write_deltaLake(events, file_format, db_name, table_name, blob_path, partition_cols, max_file_size, unique_label,
                    mode='append'):
    events.coalesce(1).writeStream \
        .format(file_format) \
        .outputMode(mode) \
        .option("checkpointLocation", f"{blob_path}/{db_name}/{table_name}/{unique_label}/_checkpoints/") \
        .option("mergeSchema", 'true') \
        .option("delta.targetFileSize", max_file_size) \
        .option("path", f'{blob_path}/{db_name}/{table_name}') \
        .partitionBy(*partition_cols) \
        .toTable(f'{db_name}.{table_name}')



"""
Batch（批量） Kafka的处理、解析相关功能
"""


# 批量读取。根据需求，按照指定的消息时间戳的范围，一般默认是最近2日，防止kafka消息推送有延迟的情况
def read_kafka(spark_ob, kafka_servers, kafka_topics, unique_label,
               start_offsets_by_timestamp, end_offsets_by_timestamp):
    return spark_ob \
        .read \
        .format("kafka") \
        .option("kafka.bootstrap.servers", kafka_servers) \
        .option("kafka.group.id", unique_label) \
        .option("kafka.session.timeout.ms", 30000) \
        .option("subscribe", kafka_topics) \
        .option("startingTimestamp", start_offsets_by_timestamp) \
        .option("endingTimestamp", end_offsets_by_timestamp) \
        .load()


# 批量写库
def write_deltaLake(events, file_format, db_name, table_name, blob_path, partition_cols, max_file_size, unique_label,
                    mode='append'):
    events.write \
        .format(file_format) \
        .mode("append") \
        .option("checkpointLocation", f"{blob_path}/{db_name}/{table_name}/{unique_label}/_checkpoints/") \
        .option("mergeSchema", 'true') \
        .option("delta.targetFileSize", max_file_size) \
        .option("path", f'{blob_path}/{db_name}/{table_name}') \
        .saveAsTable(f'{db_name}.{table_name}', partitionBy=partition_cols, mode=mode)


"""
批量读取数据对象。用于批计算/离线计算。
"""


def read_delta_table(spark_ob, db, table, date_col, recent_days=1, limit_num=5000, limit_date='2022-12-14', test=False):
    sql = f"select * from {db}.{table} where {date_col} == CURRENT_DATE() - INTERVAL '{recent_days}' DAY "
    if test:
        print(f"select * from {db}.{table} where {date_col} == to_date('{limit_date}') limit {limit_num}")
        return spark_ob.read.table(f"{db}.{table}").where(f"{date_col} == to_date('{limit_date}') limit {limit_num}")
    else:
        sql += f"limit {limit_num}"
        return spark_ob.sql(
            f"select * from {db}.{table} where {date_col} == CURRENT_DATE() - INTERVAL '{recent_days}' DAY ")



import shutil
import subprocess

from setuptools import setup, find_packages

# 读取模块详细说明
with open("README.md", "r") as fh:
    long_description = fh.read()


# 获取该包的最新版本
def get_latest_version(package_name):
    output = subprocess.run(["pip", "index", "versions", package_name],
                            capture_output=True)
    output = output.stdout.decode('utf-8')
    if output:
        output = list(filter(lambda x: len(x) > 0, output.split('\n')))
        latest_version = output[-1].split(':')[1].strip()
        return latest_version
    else:
        return None


# 拆分版本号
def split_version(v):
    return v.split('.')


PACKAGE_NAME = 'TMFuncs'
# 获取该包的最新版本
try:
    current_version = get_latest_version(PACKAGE_NAME)
    v1, v2, v3 = split_version(current_version)
except Exception as _:
    v1, v2, v3 = 0, 0, 0

# 清空dist的原有包
shutil.rmtree('dist')

# 构建新的包并生成到dist目录
setup(name=PACKAGE_NAME,
      version=f'{int(v1)}.{int(v2)}.{int(v3) + 1}',  # 默认就是v3基础上 +1
      description='Databricks common functions, written by TrueMetrics.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://www.truemetrics.cn/',
      author='TSO',
      author_email='tlsong@truemetrics.cn',
      packages=find_packages(),
      classifiers=[
          "Programming Language :: Python :: 3.9",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      )
